/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  DeadFTP
 *  Copyright (C) 2000-2001 Brandon Lees
 */

#ifndef _FXP_H_DEAD_
#define _FXP_H_DEAD_

#include <pthread.h>

#define FXP_STATUS_IDLE	     0
#define FXP_STATUS_USE_OTHER 1
#define FXP_STATUS_HOSTNAME  2
#define FXP_STATUS_START     3
#define FXP_STATUS_BUSY      4
#define FXP_STATUS_OK        5
#define FXP_STATUS_PORT      6
#define FXP_STATUS_GO        7
#define FXP_STATUS_STOP      8
#define FXP_STATUS_TRANS     9
#define FXP_STATUS_DELETE    10

#define FXP_FLAG_DIRTY       (1 << 0)

#define FXP_SET_DIRTY(fxp)    ((fxp)->data->flags |= FXP_FLAG_DIRTY)
#define FXP_DIRTY(fxp)	      ((fxp)->data->flags & FXP_FLAG_DIRTY)
#define FXP_CLEAN(fxp)	      ((fxp)->data->flags &= ~FXP_FLAG_DIRTY)
#define FXP_STATUS(fxp)	      ((fxp)->data->status)

#define FXP_SHM_SIZE	      4096

typedef struct fxp {
  pthread_mutex_t lock;

  pid_t master_pid;
  pid_t slave_pid;

  unsigned long address;
  int port;

  int status;
  unsigned long flags;
  int value;

  char master_hostname[SMALL_BUFFER];
  char slave_hostname[SMALL_BUFFER];
} fxp_t;

typedef struct fxp_desc {
  int shm_size;
  int shm_id;
  unsigned char *shm_area;

  int slave;
  int refcount;

  fxp_t *data;
} fxp_desc_t;

void init_fxp(void);
fxp_desc_t *fxp_get(void);
fxp_desc_t *fxp_attach(int shm_id);
void fxp_set_status(fxp_desc_t *, unsigned long status);
gboolean fxp_start_transfer(queue_item_t * q);
void fxp_delete(queue_item_t *);

#endif /* _FXP_H_DEAD_ */
