/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  DeadFTP
 *  Copyright (C) 2000-2001 Brandon Lees
 */

#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>
#include <glade/glade.h>
#include <gnome-xml/tree.h>
#include <gnome-xml/parser.h>

#include "globals.h"
#include "main.h"
#include "preferences.h"
#include "local.h"
#include "layout.h"
#include "view.h"
#include "remote.h"

#include "socket.h"
#include "proxy.h"

prefs_t prefs;
Preferences deadftp_prefs;

/* Local functions */
static void preferences_load_dialog(void);
static void preferences_load_values(void);

/* Callbacks in this file */
void on_preferences1_activate(GtkMenuItem * menuitem, gpointer user_data);
void on_color_remote_color_set(GnomeColorPicker * gnomecolorpicker, guint arg1,
			       guint arg2, guint arg3, guint arg4,
			       gpointer user_data);
void on_color_local_color_set(GnomeColorPicker * gnomecolorpicker, guint arg1,
			      guint arg2, guint arg3, guint arg4,
			      gpointer user_data);
void on_color_info_color_set(GnomeColorPicker * gnomecolorpicker, guint arg1,
			     guint arg2, guint arg3, guint arg4, gpointer user_data);
void on_color_error_color_set(GnomeColorPicker * gnomecolorpicker, guint arg1,
			      guint arg2, guint arg3, guint arg4,
			      gpointer user_data);
void on_remote_add_clicked(GtkButton * button, gpointer user_data);
void on_local_add_clicked(GtkButton * button, gpointer user_data);
void on_local_remove_clicked(GtkButton * button, gpointer user_data);
void on_remote_remove_clicked(GtkButton * button, gpointer user_data);
void on_queue_remove_clicked(GtkButton * button, gpointer user_data);
void on_queue_add_clicked(GtkButton * button, gpointer user_data);
void on_use_socks_auth_toggled(GtkToggleButton * togglebutton, gpointer user_data);
void on_use_socks5_toggled(GtkToggleButton * togglebutton, gpointer user_data);
void on_properties_apply(GnomePropertyBox * propertybox, gint page_num,
			 gpointer user_data);
void on_preferences_changed(void);
void on_properties_help(GnomePropertyBox * box);
void on_bind_ip_toggled(GtkToggleButton * button, gpointer data);
void on_anti_idle_toggled(GtkToggleButton * button, gpointer data);
void on_limit_port_toggled(GtkToggleButton * button, gpointer data);
void on_proxy_type_activate(GtkMenuItem * item, gpointer data);
void on_sort_field_select_child(GtkList * list, GtkWidget * child, gpointer data);

/* Local function */
static int option_menu_get_index(GtkOptionMenu * menu);

void on_preferences1_activate(GtkMenuItem * menuitem, gpointer user_data)
{
  preferences_load_dialog();
}

static void preferences_load_dialog()
{
  GladeXML *xml;
  GList *iterator;
  GtkMenu *menu;

  xml = glade_xml_new(deadftp.gladefile, "properties");
  if (!xml) {
    g_error(_("Couldn't load the interface!"));
    return;
  }

  prefs.dialog = glade_xml_get_widget(xml, "properties");

  STORE_WIDGET(prefs, anonymous_email);
  STORE_WIDGET(prefs, download_dir);
  STORE_WIDGET(prefs, log_filename);
  STORE_WIDGET(prefs, log_disk);

  STORE_WIDGET(prefs, timeout);
  STORE_WIDGET(prefs, retry_delay);
  STORE_WIDGET(prefs, scrollback);
  STORE_WIDGET(prefs, retry_count);

  STORE_WIDGET(prefs, anti_idle);
  STORE_WIDGET(prefs, anti_idle_time);
  STORE_WIDGET(prefs, popup_login_message);
  STORE_WIDGET(prefs, clipboard_monitor);
  STORE_WIDGET(prefs, selection_monitor);

  STORE_WIDGET(prefs, list_default);
  STORE_WIDGET(prefs, list_show_hidden);
  STORE_WIDGET(prefs, list_custom);
  STORE_WIDGET(prefs, list_custom_arg);

  STORE_WIDGET(prefs, color_local);
  STORE_WIDGET(prefs, color_remote);
  STORE_WIDGET(prefs, color_info);
  STORE_WIDGET(prefs, color_error);

  STORE_WIDGET(prefs, resume_u_small);
  STORE_WIDGET(prefs, resume_u_equal);
  STORE_WIDGET(prefs, resume_u_big);
  STORE_WIDGET(prefs, resume_d_small);
  STORE_WIDGET(prefs, resume_d_equal);
  STORE_WIDGET(prefs, resume_d_big);

  STORE_WIDGET(prefs, directory_possible);
  STORE_WIDGET(prefs, directory_displayed);
  STORE_WIDGET(prefs, local_sort_field);
  STORE_WIDGET(prefs, local_sort_asc);
  STORE_WIDGET(prefs, local_sort_dsc);
  STORE_WIDGET(prefs, remote_sort_field);
  STORE_WIDGET(prefs, remote_sort_asc);
  STORE_WIDGET(prefs, remote_sort_dsc);

  STORE_WIDGET(prefs, queue_possible);
  STORE_WIDGET(prefs, queue_displayed);

  STORE_WIDGET(prefs, proxy_type);
  STORE_WIDGET(prefs, proxy_host);
  STORE_WIDGET(prefs, proxy_port);
  STORE_WIDGET(prefs, proxy_username);
  STORE_WIDGET(prefs, proxy_password);
  STORE_WIDGET(prefs, proxy_use_pasv);
  STORE_WIDGET(prefs, limit_port);
  STORE_WIDGET(prefs, limit_port_low);
  STORE_WIDGET(prefs, limit_port_high);
  STORE_WIDGET(prefs, bind_ip);
  STORE_WIDGET(prefs, bind_ip_address);

  preferences_load_values();

  glade_xml_signal_autoconnect(xml);

  /* hook up signals for resume option menus */
  menu = GTK_MENU(gtk_option_menu_get_menu(GTK_OPTION_MENU(prefs.resume_u_small)));
  for (iterator = GTK_MENU_SHELL(menu)->children; iterator;
       iterator = g_list_next(iterator)) {
    gtk_signal_connect(GTK_OBJECT(iterator->data), "activate",
		       GTK_SIGNAL_FUNC(on_preferences_changed), NULL);
  }

  menu = GTK_MENU(gtk_option_menu_get_menu(GTK_OPTION_MENU(prefs.resume_u_equal)));
  for (iterator = GTK_MENU_SHELL(menu)->children; iterator;
       iterator = g_list_next(iterator)) {
    gtk_signal_connect(GTK_OBJECT(iterator->data), "activate",
		       GTK_SIGNAL_FUNC(on_preferences_changed), NULL);
  }

  menu = GTK_MENU(gtk_option_menu_get_menu(GTK_OPTION_MENU(prefs.resume_u_big)));
  for (iterator = GTK_MENU_SHELL(menu)->children; iterator;
       iterator = g_list_next(iterator)) {
    gtk_signal_connect(GTK_OBJECT(iterator->data), "activate",
		       GTK_SIGNAL_FUNC(on_preferences_changed), NULL);
  }

  menu = GTK_MENU(gtk_option_menu_get_menu(GTK_OPTION_MENU(prefs.resume_d_small)));
  for (iterator = GTK_MENU_SHELL(menu)->children; iterator;
       iterator = g_list_next(iterator)) {
    gtk_signal_connect(GTK_OBJECT(iterator->data), "activate",
		       GTK_SIGNAL_FUNC(on_preferences_changed), NULL);
  }

  menu = GTK_MENU(gtk_option_menu_get_menu(GTK_OPTION_MENU(prefs.resume_d_equal)));
  for (iterator = GTK_MENU_SHELL(menu)->children; iterator;
       iterator = g_list_next(iterator)) {
    gtk_signal_connect(GTK_OBJECT(iterator->data), "activate",
		       GTK_SIGNAL_FUNC(on_preferences_changed), NULL);
  }

  menu = GTK_MENU(gtk_option_menu_get_menu(GTK_OPTION_MENU(prefs.resume_d_big)));
  for (iterator = GTK_MENU_SHELL(menu)->children; iterator;
       iterator = g_list_next(iterator)) {
    gtk_signal_connect(GTK_OBJECT(iterator->data), "activate",
		       GTK_SIGNAL_FUNC(on_preferences_changed), NULL);
  }

  gtk_signal_connect(GTK_OBJECT(GTK_COMBO(prefs.local_sort_field)->list),
		     "select_child", GTK_SIGNAL_FUNC(on_sort_field_select_child),
		     NULL);
  gtk_signal_connect(GTK_OBJECT(GTK_COMBO(prefs.remote_sort_field)->list),
		     "select_child", GTK_SIGNAL_FUNC(on_sort_field_select_child),
		     NULL);
  gtk_clist_set_reorderable(GTK_CLIST(prefs.directory_displayed), TRUE);
  gtk_clist_set_reorderable(GTK_CLIST(prefs.queue_displayed), TRUE);

  gtk_object_unref(GTK_OBJECT(xml));
}

static void preferences_load_values()
{
  GList *l;
  GtkWidget *w;
  int i, row;
#if NUM_DIR_LAYOUTS > NUM_QUEUE_LAYOUTS
  int layout[NUM_DIR_LAYOUTS];
#else
  int layout[NUM_QUEUE_LAYOUTS];
#endif
  char *values[1];

  gtk_entry_set_text(GTK_ENTRY(prefs.anonymous_email),
		     deadftp_prefs.anonymous_password);

  w = gnome_file_entry_gtk_entry(GNOME_FILE_ENTRY(prefs.download_dir));
  gtk_entry_set_text(GTK_ENTRY(w), deadftp_prefs.download_directory);
  gnome_file_entry_set_directory(GNOME_FILE_ENTRY(prefs.download_dir), TRUE);

  w = gnome_file_entry_gtk_entry(GNOME_FILE_ENTRY(prefs.log_filename));
  gtk_entry_set_text(GTK_ENTRY(w), deadftp_prefs.log_filename);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(prefs.log_disk),
			       deadftp_prefs.logging);

  gtk_spin_button_set_value(GTK_SPIN_BUTTON(prefs.timeout),
			    deadftp_prefs.connect_timeout);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(prefs.retry_delay),
			    deadftp_prefs.retry_delay);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(prefs.retry_count),
			    deadftp_prefs.retry_count);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(prefs.scrollback),
			    deadftp_prefs.log_scrollback);

  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(prefs.anti_idle),
			       deadftp_prefs.anti_idle);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(prefs.anti_idle_time),
			    deadftp_prefs.anti_idle_time);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(prefs.popup_login_message),
			       deadftp_prefs.login_message);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(prefs.clipboard_monitor),
			       deadftp_prefs.clipboard_monitor);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(prefs.selection_monitor),
			       deadftp_prefs.selection_monitor);

  switch (deadftp_prefs.list_method) {
  case LIST_DEFAULT:
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(prefs.list_default), TRUE);
    break;
  case LIST_SHOW_HIDDEN:
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(prefs.list_show_hidden), TRUE);
    break;
  case LIST_CUSTOM:
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(prefs.list_custom), TRUE);
    break;
  }
  gtk_entry_set_text(GTK_ENTRY(prefs.list_custom_arg), deadftp_prefs.list_command);

  gnome_color_picker_set_i16(GNOME_COLOR_PICKER(prefs.color_remote),
                             deadftp_prefs.remote.red,
			     deadftp_prefs.remote.green,
			     deadftp_prefs.remote.blue, 0);
  gnome_color_picker_set_i16(GNOME_COLOR_PICKER(prefs.color_local),
                             deadftp_prefs.local.red,
			     deadftp_prefs.local.green,
			     deadftp_prefs.local.blue, 0);
  gnome_color_picker_set_i16(GNOME_COLOR_PICKER(prefs.color_info),
                             deadftp_prefs.info.red,
			     deadftp_prefs.info.green,
			     deadftp_prefs.info.blue, 0);
  gnome_color_picker_set_i16(GNOME_COLOR_PICKER(prefs.color_error),
                             deadftp_prefs.error.red,
			     deadftp_prefs.error.green,
			     deadftp_prefs.error.blue, 0);

  for (i = 0; i < NUM_DIR_LAYOUTS; i++)
    layout[i] = -1;

  for (i = 0; i < NUM_DIR_LAYOUTS; i++) {
    if (deadftp_prefs.directory_layout[i] == -1) {
      values[0] = _(dir_layout_titles[i]);
      row = gtk_clist_append(GTK_CLIST(prefs.directory_possible), values);
      gtk_clist_set_row_data(GTK_CLIST(prefs.directory_possible), row,
			     GINT_TO_POINTER(i));
    } else {
      layout[deadftp_prefs.directory_layout[i]] = i;
    }
  }

  for (i = 0; i < NUM_DIR_LAYOUTS; i++) {
    if (layout[i] != -1) {
      values[0] = _(dir_layout_titles[layout[i]]);
      row = gtk_clist_append(GTK_CLIST(prefs.directory_displayed), values);
      gtk_clist_set_row_data(GTK_CLIST(prefs.directory_displayed), row,
			     GINT_TO_POINTER(layout[i]));
    }
  }

  if (deadftp_prefs.local_sort_order == DIR_SORT_ASC)
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(prefs.local_sort_asc), TRUE);
  else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(prefs.local_sort_dsc), TRUE);

  if (deadftp_prefs.remote_sort_order == DIR_SORT_ASC)
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(prefs.remote_sort_asc), TRUE);
  else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(prefs.remote_sort_dsc), TRUE);

  for (i = 0; i < NUM_DIR_LAYOUTS; i++) {
    w = gtk_list_item_new_with_label(_(dir_layout_titles[i]));
    gtk_object_set_data(GTK_OBJECT(w), "layout-index", GINT_TO_POINTER(i));
    gtk_widget_show(w);

    gtk_container_add(GTK_CONTAINER(GTK_COMBO(prefs.local_sort_field)->list), w);
  }
  gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(prefs.local_sort_field)->entry),
		     _(dir_layout_titles[deadftp_prefs.local_sort_field]));
  gtk_object_set_data(GTK_OBJECT(GTK_COMBO(prefs.local_sort_field)->list),
		      "layout-index",
		      GINT_TO_POINTER(deadftp_prefs.local_sort_field));

  for (i = 0; i < NUM_DIR_LAYOUTS; i++) {
    w = gtk_list_item_new_with_label(_(dir_layout_titles[i]));
    gtk_object_set_data(GTK_OBJECT(w), "layout-index", GINT_TO_POINTER(i));
    gtk_widget_show(w);

    gtk_container_add(GTK_CONTAINER(GTK_COMBO(prefs.remote_sort_field)->list), w);
  }
  gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(prefs.remote_sort_field)->entry),
		     _(dir_layout_titles[deadftp_prefs.remote_sort_field]));
  gtk_object_set_data(GTK_OBJECT(GTK_COMBO(prefs.remote_sort_field)->entry),
		      "layout-index",
		      GINT_TO_POINTER(deadftp_prefs.remote_sort_field));

  for (i = 0; i < NUM_QUEUE_LAYOUTS; i++)
    layout[i] = -1;

  for (i = 0; i < NUM_QUEUE_LAYOUTS; i++) {
    if (deadftp_prefs.queue_layout[i] == -1) {
      values[0] = _(queue_layout_titles[i]);
      row = gtk_clist_append(GTK_CLIST(prefs.queue_possible), values);
      gtk_clist_set_row_data(GTK_CLIST(prefs.queue_possible), row,
			     GINT_TO_POINTER(i));
    } else {
      layout[deadftp_prefs.queue_layout[i]] = i;
    }
  }

  for (i = 0; i < NUM_QUEUE_LAYOUTS; i++) {
    if (layout[i] != -1) {
      values[0] = _(queue_layout_titles[layout[i]]);
      row = gtk_clist_append(GTK_CLIST(prefs.queue_displayed), values);
      gtk_clist_set_row_data(GTK_CLIST(prefs.queue_displayed), row,
			     GINT_TO_POINTER(layout[i]));
    }
  }

  l = NULL;
  w = gtk_menu_new();
  for (i = 0; proxies[i].desc; i++) {
    GtkWidget *item = gtk_menu_item_new_with_label(_(proxies[i].desc));
    gtk_object_set_data(GTK_OBJECT(item), "proxy_type", GINT_TO_POINTER(i));
    gtk_signal_connect(GTK_OBJECT(item), "activate",
		       GTK_SIGNAL_FUNC(on_proxy_type_activate), GINT_TO_POINTER(i));
    gtk_menu_append(GTK_MENU(w), item);
  }
  gtk_widget_show_all(w);
  gtk_option_menu_set_menu(GTK_OPTION_MENU(prefs.proxy_type), w);
  gtk_option_menu_set_history(GTK_OPTION_MENU(prefs.proxy_type),
			      deadftp_prefs.firewall.proxy_type);
  gtk_entry_set_text(GTK_ENTRY(prefs.proxy_host),
		     deadftp_prefs.firewall.proxy_hostname);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(prefs.proxy_port),
			    deadftp_prefs.firewall.proxy_port);
  gtk_entry_set_text(GTK_ENTRY(prefs.proxy_username),
		     deadftp_prefs.firewall.proxy_username);
  gtk_entry_set_text(GTK_ENTRY(prefs.proxy_password),
		     deadftp_prefs.firewall.proxy_password);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(prefs.proxy_use_pasv),
			       deadftp_prefs.firewall.use_pasv);

  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(prefs.limit_port),
			       deadftp_prefs.firewall.limit_port_range);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(prefs.limit_port_low),
			    deadftp_prefs.firewall.limit_port_low);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(prefs.limit_port_high),
			    deadftp_prefs.firewall.limit_port_high);

  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(prefs.bind_ip),
			       deadftp_prefs.firewall.bind_ip);
  gtk_entry_set_text(GTK_ENTRY(prefs.bind_ip_address),
		     deadftp_prefs.firewall.bind_ip_address);

  gtk_option_menu_set_history(GTK_OPTION_MENU(prefs.resume_u_small),
			      deadftp_prefs.resume_upload[RESUME_COND_SMALLER]);
  gtk_option_menu_set_history(GTK_OPTION_MENU(prefs.resume_u_equal),
			      deadftp_prefs.resume_upload[RESUME_COND_EQUAL]);
  gtk_option_menu_set_history(GTK_OPTION_MENU(prefs.resume_u_big),
			      deadftp_prefs.resume_upload[RESUME_COND_BIGGER]);
  gtk_option_menu_set_history(GTK_OPTION_MENU(prefs.resume_d_small),
			      deadftp_prefs.resume_download[RESUME_COND_SMALLER]);
  gtk_option_menu_set_history(GTK_OPTION_MENU(prefs.resume_d_equal),
			      deadftp_prefs.resume_download[RESUME_COND_EQUAL]);
  gtk_option_menu_set_history(GTK_OPTION_MENU(prefs.resume_d_big),
			      deadftp_prefs.resume_download[RESUME_COND_BIGGER]);
}

void on_properties_apply(GnomePropertyBox * propertybox, gint page_num,
			 gpointer user_data)
{
  int i, col;
  gushort alpha;

  g_warning("Apply page %d\n", page_num);
  switch (page_num) {
  case 0:
    if (deadftp_prefs.anonymous_password)
      g_free(deadftp_prefs.anonymous_password);
    deadftp_prefs.anonymous_password =
      g_strdup(gtk_entry_get_text(GTK_ENTRY(prefs.anonymous_email)));
    if (deadftp_prefs.download_directory)
      g_free(deadftp_prefs.download_directory);
    deadftp_prefs.download_directory =
      gnome_file_entry_get_full_path(GNOME_FILE_ENTRY(prefs.download_dir), TRUE);
    if (deadftp_prefs.log_filename)
      g_free(deadftp_prefs.log_filename);
    deadftp_prefs.log_filename =
      gnome_file_entry_get_full_path(GNOME_FILE_ENTRY(prefs.log_filename), FALSE);
    deadftp_prefs.logging =
      gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(prefs.log_disk));
    deadftp_prefs.connect_timeout =
      gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(prefs.timeout));
    deadftp_prefs.retry_delay =
      gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(prefs.retry_delay));
    deadftp_prefs.log_scrollback =
      gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(prefs.scrollback));
    deadftp_prefs.retry_count =
      gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(prefs.retry_count));

    /* reopen the console logfile to ensure it is opened ro closed as appropriate */
    console_log_init();

    break;
  case 1:
    deadftp_prefs.anti_idle =
      gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(prefs.anti_idle));
    deadftp_prefs.anti_idle_time =
      gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(prefs.anti_idle_time));
    deadftp_prefs.login_message =
      gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(prefs.popup_login_message));
    deadftp_prefs.clipboard_monitor =
      gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(prefs.clipboard_monitor));
    deadftp_prefs.selection_monitor =
      gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(prefs.selection_monitor));
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(prefs.list_default)))
      deadftp_prefs.list_method = LIST_DEFAULT;
    else if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(prefs.list_show_hidden)))
      deadftp_prefs.list_method = LIST_SHOW_HIDDEN;
    else {
      deadftp_prefs.list_method = LIST_CUSTOM;
      if (deadftp_prefs.list_command)
	g_free(deadftp_prefs.list_command);
      deadftp_prefs.list_command =
	g_strdup(gtk_entry_get_text(GTK_ENTRY(prefs.list_custom_arg)));
    }
    break;
  case 2:
    gnome_color_picker_get_i16(GNOME_COLOR_PICKER(prefs.color_remote),
                               &deadftp_prefs.remote.red,
                               &deadftp_prefs.remote.green,
                               &deadftp_prefs.remote.blue, &alpha);

    gnome_color_picker_get_i16(GNOME_COLOR_PICKER(prefs.color_error),
                               &deadftp_prefs.error.red, &deadftp_prefs.error.green,
                               &deadftp_prefs.error.blue, &alpha);

    gnome_color_picker_get_i16(GNOME_COLOR_PICKER(prefs.color_info),
                               &deadftp_prefs.info.red, &deadftp_prefs.info.green,
                               &deadftp_prefs.info.blue, &alpha);

    gnome_color_picker_get_i16(GNOME_COLOR_PICKER(prefs.color_local),
                               &deadftp_prefs.local.red, &deadftp_prefs.local.green,
                               &deadftp_prefs.local.blue, &alpha);
    break;
  case 3:
    /* I don't really like this method, it's somewhat hackish and only works
     * because the items in the resume option menu are in the same order they're
     * in resume_t but I can't really think of a way I like better */
    deadftp_prefs.resume_upload[RESUME_COND_SMALLER] =
      option_menu_get_index(GTK_OPTION_MENU(prefs.resume_u_small));
    deadftp_prefs.resume_upload[RESUME_COND_EQUAL] =
      option_menu_get_index(GTK_OPTION_MENU(prefs.resume_u_equal));
    deadftp_prefs.resume_upload[RESUME_COND_BIGGER] =
      option_menu_get_index(GTK_OPTION_MENU(prefs.resume_u_big));
    deadftp_prefs.resume_download[RESUME_COND_SMALLER] =
      option_menu_get_index(GTK_OPTION_MENU(prefs.resume_d_small));
    deadftp_prefs.resume_download[RESUME_COND_EQUAL] =
      option_menu_get_index(GTK_OPTION_MENU(prefs.resume_d_equal));
    deadftp_prefs.resume_download[RESUME_COND_BIGGER] =
      option_menu_get_index(GTK_OPTION_MENU(prefs.resume_d_big));
    break;
  case 4:
    for (i = 0; i < NUM_DIR_LAYOUTS; i++)
      deadftp_prefs.directory_layout[i] = -1;

    for (i = 0; i < GTK_CLIST(prefs.directory_displayed)->rows; i++) {
      col =
	GPOINTER_TO_INT(gtk_clist_get_row_data
			(GTK_CLIST(prefs.directory_displayed), i));
      deadftp_prefs.directory_layout[col] = i;
    }

    deadftp_prefs.local_sort_order =
      gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(prefs.local_sort_dsc));
    deadftp_prefs.local_sort_field =
      GPOINTER_TO_INT(gtk_object_get_data
		      (GTK_OBJECT(GTK_COMBO(prefs.local_sort_field)->list),
		       "layout-index"));

    deadftp_prefs.remote_sort_order =
      gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(prefs.remote_sort_dsc));
    deadftp_prefs.remote_sort_field =
      GPOINTER_TO_INT(gtk_object_get_data
		      (GTK_OBJECT(GTK_COMBO(prefs.remote_sort_field)->list),
		       "layout-index"));

    gtk_clist_clear(GTK_CLIST(deadftp.local));
    gtk_clist_clear(GTK_CLIST(deadftp.remote));

    init_local_dir_layout();
    init_remote_dir_layout();

    on_local_refresh();
    on_remote_refresh();
    break;
  case 5:
    for (i = 0; i < NUM_QUEUE_LAYOUTS; i++)
      deadftp_prefs.queue_layout[i] = -1;

    for (i = 0; i < GTK_CLIST(prefs.queue_displayed)->rows; i++) {
      col =
	GPOINTER_TO_INT(gtk_clist_get_row_data(GTK_CLIST(prefs.queue_displayed), i));
      deadftp_prefs.queue_layout[col] = i;
    }

    gtk_clist_clear(GTK_CLIST(deadftp.queue));
    init_queue_layout();
    redisplay_queue();
    break;
  case 6:
    deadftp_prefs.firewall.proxy_type =
      GPOINTER_TO_INT(gtk_object_get_data
		      (GTK_OBJECT
		       (gtk_menu_get_active
			(GTK_MENU
			 (gtk_option_menu_get_menu
			  (GTK_OPTION_MENU(prefs.proxy_type))))), "proxy_type"));
    if (!PROXY_FLAG(deadftp_prefs.firewall.proxy_type, PROXY_FLAG_DISABLE))
      deadftp_prefs.firewall.use_proxy = TRUE;
    else
      deadftp_prefs.firewall.use_proxy = FALSE;

    if (deadftp_prefs.firewall.proxy_hostname)
      g_free(deadftp_prefs.firewall.proxy_hostname);
    /* gtk_entry_get_text is deprecated, use gtk_editable_get_chars instead */
    deadftp_prefs.firewall.proxy_hostname =
      gtk_editable_get_chars(GTK_EDITABLE(prefs.proxy_host), 0, -1);
    deadftp_prefs.firewall.proxy_port =
      gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(prefs.proxy_port));
    if (deadftp_prefs.firewall.proxy_username)
      g_free(deadftp_prefs.firewall.proxy_username);
    deadftp_prefs.firewall.proxy_username =
      gtk_editable_get_chars(GTK_EDITABLE(prefs.proxy_username), 0, -1);
    if (deadftp_prefs.firewall.proxy_password)
      g_free(deadftp_prefs.firewall.proxy_password);
    deadftp_prefs.firewall.proxy_password =
      gtk_editable_get_chars(GTK_EDITABLE(prefs.proxy_password), 0, -1);
    deadftp_prefs.firewall.use_pasv =
      gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(prefs.proxy_use_pasv));

    deadftp_prefs.firewall.limit_port_range =
      gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(prefs.limit_port));
    deadftp_prefs.firewall.limit_port_low =
      gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(prefs.limit_port_low));
    deadftp_prefs.firewall.limit_port_high =
      gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(prefs.limit_port_high));

    deadftp_prefs.firewall.bind_ip =
      gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(prefs.bind_ip));
    if (deadftp_prefs.firewall.bind_ip_address)
      g_free(deadftp_prefs.firewall.bind_ip_address);
    deadftp_prefs.firewall.bind_ip_address =
      g_strdup(gtk_entry_get_text(GTK_ENTRY(prefs.bind_ip_address)));
    break;
  }

  save_preferences();
}

void on_color_remote_color_set(GnomeColorPicker * gnomecolorpicker, guint arg1,
			       guint arg2, guint arg3, guint arg4,
			       gpointer user_data)
{
  /* Set the ColorPicker color to the new color */
  gnome_color_picker_set_i16(gnomecolorpicker, arg1, arg2, arg3, arg4);

  on_preferences_changed();
}

void on_color_local_color_set(GnomeColorPicker * gnomecolorpicker, guint arg1,
			      guint arg2, guint arg3, guint arg4, gpointer user_data)
{
  /* Set the ColorPicker color to the new color */
  gnome_color_picker_set_i16(gnomecolorpicker, arg1, arg2, arg3, arg4);

  on_preferences_changed();
}

void on_color_info_color_set(GnomeColorPicker * gnomecolorpicker, guint arg1,
			     guint arg2, guint arg3, guint arg4, gpointer user_data)
{
  /* Set the ColorPicker color to the new color */
  gnome_color_picker_set_i16(gnomecolorpicker, arg1, arg2, arg3, arg4);

  on_preferences_changed();
}

void on_color_error_color_set(GnomeColorPicker * gnomecolorpicker, guint arg1,
			      guint arg2, guint arg3, guint arg4, gpointer user_data)
{
  /* Set the ColorPicker color to the new color */
  gnome_color_picker_set_i16(gnomecolorpicker, arg1, arg2, arg3, arg4);

  on_preferences_changed();
}

void on_preferences_changed(void)
{
  gnome_property_box_changed(GNOME_PROPERTY_BOX(prefs.dialog));
}

void on_properties_help(GnomePropertyBox * box)
{
  GnomeHelpMenuEntry help_entry = { "deadftp", "prefs.html" };
  gnome_help_display(NULL, &help_entry);
}

void load_preferences()
{
  gchar *str, **col;
  int i;

  deadftp_prefs.anonymous_password =
    gnome_config_get_string("Options/anonymous-password=deadftp@");
  str = g_strdup_printf("Options/download-directory=%s", g_get_home_dir());
  deadftp_prefs.download_directory = gnome_config_get_string(str);
  g_free(str);
  deadftp_prefs.log_filename =
    gnome_config_get_string("Options/log-filename=ftplog.txt");
  deadftp_prefs.logging = gnome_config_get_bool("Options/logging=false");

  deadftp_prefs.connect_timeout = gnome_config_get_int("Options/timeout=300");
  deadftp_prefs.retry_delay = gnome_config_get_int("Options/retry-delay=60");
  deadftp_prefs.retry_count = gnome_config_get_int("Options/retry-count=10");
  deadftp_prefs.log_scrollback = gnome_config_get_int("Options/scrollback=200");

  deadftp_prefs.anti_idle = gnome_config_get_bool("Options/keepalive=false");
  deadftp_prefs.anti_idle_time = gnome_config_get_int("Options/keepalive-time=120");
  deadftp_prefs.login_message = gnome_config_get_bool("Options/login-message=true");
  deadftp_prefs.selection_monitor =
    gnome_config_get_bool("Options/selection-monitor=false");
  deadftp_prefs.clipboard_monitor =
    gnome_config_get_bool("Options/clipboard-monitor=true");

  deadftp_prefs.list_method = gnome_config_get_int("Options/list-method=0");
  deadftp_prefs.list_command = gnome_config_get_string("Options/list-command=");

  deadftp_prefs.finish_action = gnome_config_get_int("Options/finish-action=0");

  str = gnome_config_get_string("Color/remote=#0000bfff0000");
  gdk_color_parse(str, &deadftp_prefs.remote);
  g_free(str);

  str = gnome_config_get_string("Color/local=#00000000bfff");
  gdk_color_parse(str, &deadftp_prefs.local);
  g_free(str);

  str = gnome_config_get_string("Color/info=#bfffbfff0000");
  gdk_color_parse(str, &deadftp_prefs.info);
  g_free(str);

  str = gnome_config_get_string("Color/error=#bfff00000000");
  gdk_color_parse(str, &deadftp_prefs.error);
  g_free(str);

  /* dir layout string = file,len,mode,owner,group,date */
  str = gnome_config_get_string("Layout/Directory=0,1,2,-1,-1,3");
  col = g_strsplit(str, ",", 6);
  for (i = 0; i < NUM_DIR_LAYOUTS; i++)
    deadftp_prefs.directory_layout[i] = atoi(col[i]);
  g_strfreev(col);
  g_free(str);
  deadftp_prefs.local_sort_field = gnome_config_get_int("Layout/local-sort-field=0");
  deadftp_prefs.local_sort_order =
    gnome_config_get_bool("Layout/local-sort-order=false");
  deadftp_prefs.local_auto_size =
    gnome_config_get_bool("Layout/local-auto-size=true");

  deadftp_prefs.remote_sort_field =
    gnome_config_get_int("Layout/remote-sort-field=0");
  deadftp_prefs.remote_sort_order =
    gnome_config_get_bool("Layout/remote-sort-order=false");
  deadftp_prefs.remote_auto_size =
    gnome_config_get_bool("Layout/remote-auto-size=true");

  /* queue layout str =
   * local_file,local_dir,remote_file,remote_dir,hostname,port,username,size,status,
   * remote_path,local_path,url,direction */

  str = gnome_config_get_string("Layout/queue=0,-1,-1,-1,-1,-1,-1,3,2,4,-1,-1,1");
  col = g_strsplit(str, ",", 13);
  for (i = 0; i < NUM_QUEUE_LAYOUTS; i++)
    deadftp_prefs.queue_layout[i] = atoi(col[i]);
  g_strfreev(col);
  g_free(str);

  deadftp_prefs.firewall.proxy_type = gnome_config_get_int("Firewall/proxy-type=0");
  if (!PROXY_FLAG(deadftp_prefs.firewall.proxy_type, PROXY_FLAG_DISABLE))
    deadftp_prefs.firewall.use_proxy = TRUE;
  else
    deadftp_prefs.firewall.use_proxy = FALSE;
  deadftp_prefs.firewall.proxy_hostname =
    gnome_config_get_string("Firewall/proxy-hostname=");
  deadftp_prefs.firewall.proxy_port =
    gnome_config_get_int("Firewall/proxy-port=1080");
  deadftp_prefs.firewall.proxy_username =
    gnome_config_get_string("Firewall/proxy-username=");
  deadftp_prefs.firewall.proxy_password =
    gnome_config_private_get_string("Firewall/proxy-password=");
  deadftp_prefs.firewall.use_pasv = gnome_config_get_bool("Firewall/use-pasv=false");

  deadftp_prefs.firewall.limit_port_range =
    gnome_config_get_bool("Firewall/limit-port-range=false");
  deadftp_prefs.firewall.limit_port_low =
    gnome_config_get_int("Firewall/limit-port-low=1024");
  deadftp_prefs.firewall.limit_port_high =
    gnome_config_get_int("Firewall/limit-port-high=2048");

  deadftp_prefs.firewall.bind_ip = gnome_config_get_bool("Firewall/bind-ip=false");
  deadftp_prefs.firewall.bind_ip_address =
    gnome_config_get_string("Firewall/bind-ip-address=");

  gnome_config_get_vector("History/hostname", &i, &col);
  for (i--; i >= 0; i--)
    deadftp_prefs.hostname_history =
      g_list_append(deadftp_prefs.hostname_history, g_strdup(col[i]));
  g_free(col);

  gnome_config_get_vector("History/port", &i, &col);
  for (i--; i >= 0; i--)
    deadftp_prefs.port_history =
      g_list_append(deadftp_prefs.port_history, g_strdup(col[i]));
  g_free(col);

  gnome_config_get_vector("History/username", &i, &col);
  for (i--; i >= 0; i--)
    deadftp_prefs.username_history =
      g_list_append(deadftp_prefs.username_history, g_strdup(col[i]));
  g_free(col);

  deadftp_prefs.resume_upload[RESUME_COND_SMALLER] =
    gnome_config_get_int("resume-upload/smaller=0");
  deadftp_prefs.resume_upload[RESUME_COND_EQUAL] =
    gnome_config_get_int("resume-upload/equal=0");
  deadftp_prefs.resume_upload[RESUME_COND_BIGGER] =
    gnome_config_get_int("resume-upload/bigger=0");
  deadftp_prefs.resume_download[RESUME_COND_SMALLER] =
    gnome_config_get_int("resume-download/smaller=0");
  deadftp_prefs.resume_download[RESUME_COND_EQUAL] =
    gnome_config_get_int("resume-download/equal=0");
  deadftp_prefs.resume_download[RESUME_COND_BIGGER] =
    gnome_config_get_int("resume-download/bigger=0");

  deadftp_prefs.resume_timeout = gnome_config_get_bool("resume-timeout=true");
  deadftp_prefs.resume_timeout_time = gnome_config_get_int("resume-timeout-time=30");
}

void save_preferences()
{
  gchar *str, buffer[SMALL_BUFFER], **ptr;
  int i;
  GList *l;

  gnome_config_set_string("Options/anonymous-password",
			  deadftp_prefs.anonymous_password);
  gnome_config_set_string("Options/download-directory",
			  deadftp_prefs.download_directory);
  gnome_config_set_string("Options/log-filename", deadftp_prefs.log_filename);
  gnome_config_set_bool("Options/logging", deadftp_prefs.logging);

  gnome_config_set_int("Options/timeout", deadftp_prefs.connect_timeout);
  gnome_config_set_int("Options/retry-delay", deadftp_prefs.retry_delay);
  gnome_config_set_int("Options/retry-count", deadftp_prefs.retry_count);
  gnome_config_set_int("Options/log-scrollback", deadftp_prefs.log_scrollback);

  gnome_config_set_bool("Options/keepalive", deadftp_prefs.anti_idle);
  gnome_config_set_int("Options/keepalive-time", deadftp_prefs.anti_idle_time);
  gnome_config_set_bool("Options/login-message", deadftp_prefs.login_message);
  gnome_config_set_bool("Options/clipboard-monitor",
			deadftp_prefs.clipboard_monitor);
  gnome_config_set_bool("Options/selection-monitor",
			deadftp_prefs.selection_monitor);

  gnome_config_set_int("Options/list-method", deadftp_prefs.list_method);
  gnome_config_set_string("Options/list-command", deadftp_prefs.list_command);

  gnome_config_set_int("Options/finish-action", deadftp_prefs.finish_action);

  str =
    g_strdup_printf("#%04x%04x%04x", deadftp_prefs.error.red,
		    deadftp_prefs.error.green, deadftp_prefs.error.blue);
  gnome_config_set_string("Color/error", str);
  g_free(str);

  str =
    g_strdup_printf("#%04x%04x%04x", deadftp_prefs.info.red,
		    deadftp_prefs.info.green, deadftp_prefs.info.blue);
  gnome_config_set_string("Color/info", str);
  g_free(str);

  str =
    g_strdup_printf("#%04x%04x%04x", deadftp_prefs.local.red,
		    deadftp_prefs.local.green, deadftp_prefs.local.blue);
  gnome_config_set_string("Color/local", str);
  g_free(str);

  str =
    g_strdup_printf("#%04x%04x%04x", deadftp_prefs.remote.red,
		    deadftp_prefs.remote.green, deadftp_prefs.remote.blue);
  gnome_config_set_string("Color/remote", str);
  g_free(str);

  *buffer = '\0';
  for (i = 0; i < (NUM_DIR_LAYOUTS - 1); i++)
    g_snprintf(buffer + strlen(buffer), SMALL_BUFFER - strlen(buffer), "%d,",
	       deadftp_prefs.directory_layout[i]);
  g_snprintf(buffer + strlen(buffer), SMALL_BUFFER - strlen(buffer), "%d",
	     deadftp_prefs.directory_layout[i]);
  gnome_config_set_string("Layout/Directory", buffer);

  *buffer = '\0';
  for (i = 0; i < (NUM_QUEUE_LAYOUTS - 1); i++)
    g_snprintf(buffer + strlen(buffer), SMALL_BUFFER - strlen(buffer), "%d,",
	       deadftp_prefs.queue_layout[i]);
  g_snprintf(buffer + strlen(buffer), SMALL_BUFFER - strlen(buffer), "%d",
	     deadftp_prefs.queue_layout[i]);
  gnome_config_set_string("Layout/queue", buffer);

  gnome_config_set_int("Firewall/proxy-type", deadftp_prefs.firewall.proxy_type);
  gnome_config_set_string("Firewall/proxy-hostname",
			  deadftp_prefs.firewall.proxy_hostname);
  gnome_config_set_int("Firewall/proxy-port", deadftp_prefs.firewall.proxy_port);
  gnome_config_set_string("Firewall/proxy-username",
			  deadftp_prefs.firewall.proxy_username);
  gnome_config_private_set_string("Firewall/proxy-password",
				  deadftp_prefs.firewall.proxy_password);
  gnome_config_set_bool("Firewall/use-pasv", deadftp_prefs.firewall.use_pasv);

  gnome_config_set_bool("Firewall/limit-port-range",
			deadftp_prefs.firewall.limit_port_range);
  gnome_config_set_int("Firewall/limit-port-low",
		       deadftp_prefs.firewall.limit_port_low);
  gnome_config_set_int("Firewall/limit-port-high",
		       deadftp_prefs.firewall.limit_port_high);

  gnome_config_set_bool("Firewall/bind-ip", deadftp_prefs.firewall.bind_ip);
  gnome_config_set_string("Firewall/bind-ip-address",
			  deadftp_prefs.firewall.bind_ip_address);

  ptr = g_new(gchar *, g_list_length(deadftp_prefs.hostname_history));
  for (l = deadftp_prefs.hostname_history, i = 0; l; l = g_list_next(l), i++)
    ptr[i] = (gchar *) l->data;
  gnome_config_set_vector("History/hostname", i, (const char **) ptr);
  g_free(ptr);

  ptr = g_new(gchar *, g_list_length(deadftp_prefs.port_history));
  for (l = deadftp_prefs.port_history, i = 0; l; l = g_list_next(l), i++)
    ptr[i] = (gchar *) l->data;
  gnome_config_set_vector("History/port", i, (const char **) ptr);
  g_free(ptr);

  ptr = g_new(gchar *, g_list_length(deadftp_prefs.username_history));
  for (l = deadftp_prefs.username_history, i = 0; l; l = g_list_next(l), i++)
    ptr[i] = (gchar *) l->data;
  gnome_config_set_vector("History/username", i, (const char **) ptr);
  g_free(ptr);

  gnome_config_set_int("resume-upload/smaller",
		       deadftp_prefs.resume_upload[RESUME_COND_SMALLER]);
  gnome_config_set_int("resume-upload/equal",
		       deadftp_prefs.resume_upload[RESUME_COND_EQUAL]);
  gnome_config_set_int("resume-upload/bigger",
		       deadftp_prefs.resume_upload[RESUME_COND_BIGGER]);
  gnome_config_set_int("resume-download/smaller",
		       deadftp_prefs.resume_download[RESUME_COND_SMALLER]);
  gnome_config_set_int("resume-download/equal",
		       deadftp_prefs.resume_download[RESUME_COND_EQUAL]);
  gnome_config_set_int("resume-download/bigger",
		       deadftp_prefs.resume_download[RESUME_COND_BIGGER]);

  gnome_config_set_bool("resume-timeout", deadftp_prefs.resume_timeout);
  gnome_config_set_int("resume-timeout-time", deadftp_prefs.resume_timeout_time);

  gnome_config_sync();
}

void on_bind_ip_toggled(GtkToggleButton * button, gpointer data)
{
  gtk_widget_set_sensitive(GTK_WIDGET(prefs.bind_ip_address),
			   gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button)));

  on_preferences_changed();
}

void on_anti_idle_toggled(GtkToggleButton * button, gpointer data)
{
  gtk_widget_set_sensitive(GTK_WIDGET(prefs.anti_idle_time),
			   gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button)));

  on_preferences_changed();
}

void on_limit_port_toggled(GtkToggleButton * button, gpointer data)
{
  gboolean active = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button));
  gtk_widget_set_sensitive(GTK_WIDGET(prefs.limit_port_low), active);
  gtk_widget_set_sensitive(GTK_WIDGET(prefs.limit_port_high), active);

  on_preferences_changed();
}

void on_proxy_type_activate(GtkMenuItem * item, gpointer data)
{
  gint type = GPOINTER_TO_INT(data);
  g_warning("got proxy type %d\n", type);

  g_return_if_fail(proxies[type].desc != NULL);

  if (PROXY_FLAG(type, PROXY_FLAG_DISABLE)) {
    gtk_widget_set_sensitive(GTK_WIDGET(prefs.proxy_host), FALSE);
    gtk_widget_set_sensitive(GTK_WIDGET(prefs.proxy_port), FALSE);
  } else {
    gtk_widget_set_sensitive(GTK_WIDGET(prefs.proxy_host), TRUE);
    gtk_widget_set_sensitive(GTK_WIDGET(prefs.proxy_port), TRUE);
  }
  if (PROXY_FLAG(type, PROXY_FLAG_AUTH) && !PROXY_FLAG(type, PROXY_FLAG_DISABLE)) {
    gtk_widget_set_sensitive(GTK_WIDGET(prefs.proxy_username), TRUE);
    gtk_widget_set_sensitive(GTK_WIDGET(prefs.proxy_password), TRUE);
  } else {
    gtk_widget_set_sensitive(GTK_WIDGET(prefs.proxy_username), FALSE);
    gtk_widget_set_sensitive(GTK_WIDGET(prefs.proxy_password), FALSE);
  }
  if (PROXY_FLAG(type, PROXY_FLAG_PORT)) {
    gtk_widget_set_sensitive(GTK_WIDGET(prefs.proxy_use_pasv), TRUE);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(prefs.proxy_use_pasv),
				 deadftp_prefs.firewall.use_pasv);
  } else {
    gtk_widget_set_sensitive(GTK_WIDGET(prefs.proxy_use_pasv), FALSE);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(prefs.proxy_use_pasv), TRUE);
  }

  on_preferences_changed();
}

static int option_menu_get_index(GtkOptionMenu * optmenu)
{
  GtkMenu *menu;
  GtkMenuItem *item;
  GList *iterator;
  int index;

  g_return_val_if_fail(optmenu != NULL, -1);

  menu = GTK_MENU(gtk_option_menu_get_menu(optmenu));
  iterator = GTK_MENU_SHELL(menu)->children;
  item = GTK_MENU_ITEM(gtk_menu_get_active(menu));

  index = 0;
  while (iterator && iterator->data != item) {
    index++;
    iterator = g_list_next(iterator);
  }

  if (!iterator) {
    /* The selected item was not found.  Either no selection or an error
     * occurred. */
    return -1;
  }

  g_warning("Got index %d\n", index);
  return index;
}

void on_sort_field_select_child(GtkList * list, GtkWidget * child, gpointer data)
{
  int index;

  index = GPOINTER_TO_INT(gtk_object_get_data(GTK_OBJECT(child), "layout-index"));
  gtk_object_set_data(GTK_OBJECT(list), "layout-index", GINT_TO_POINTER(index));
}

void on_directory_add_clicked(GtkButton * button, gpointer data)
{
  int index, old_row, new_row;
  char *values[1];

  if (!GTK_CLIST(prefs.directory_possible)->selection)
    return;

  old_row = GPOINTER_TO_INT(GTK_CLIST(prefs.directory_possible)->selection->data);
  index =
    GPOINTER_TO_INT(gtk_clist_get_row_data
		    (GTK_CLIST(prefs.directory_possible), old_row));
  gtk_clist_get_text(GTK_CLIST(prefs.directory_possible), old_row, 0, &values[0]);

  new_row = gtk_clist_append(GTK_CLIST(prefs.directory_displayed), values);
  gtk_clist_set_row_data(GTK_CLIST(prefs.directory_displayed), new_row,
			 GINT_TO_POINTER(index));

  gtk_clist_remove(GTK_CLIST(prefs.directory_possible), old_row);

  on_preferences_changed();
}

void on_directory_remove_clicked(GtkButton * button, gpointer data)
{
  int index, old_row, new_row;
  char *values[1];

  if (!GTK_CLIST(prefs.directory_displayed)->selection)
    return;

  old_row = GPOINTER_TO_INT(GTK_CLIST(prefs.directory_displayed)->selection->data);
  index =
    GPOINTER_TO_INT(gtk_clist_get_row_data
		    (GTK_CLIST(prefs.directory_displayed), old_row));
  gtk_clist_get_text(GTK_CLIST(prefs.directory_displayed), old_row, 0, &values[0]);

  new_row = gtk_clist_insert(GTK_CLIST(prefs.directory_possible), index, values);
  gtk_clist_set_row_data(GTK_CLIST(prefs.directory_possible), new_row,
			 GINT_TO_POINTER(index));

  gtk_clist_remove(GTK_CLIST(prefs.directory_displayed), old_row);

  on_preferences_changed();
}

void on_directory_up_clicked(GtkButton * button, gpointer data)
{
  int row;

  if (!GTK_CLIST(prefs.directory_displayed)->selection)
    return;

  row = GPOINTER_TO_INT(GTK_CLIST(prefs.directory_displayed)->selection->data);
  /* Don't try to move the row if it's already at the top */
  if (row == 0)
    return;

  gtk_clist_row_move(GTK_CLIST(prefs.directory_displayed), row, row - 1);

  on_preferences_changed();
}

void on_directory_down_clicked(GtkButton * button, gpointer data)
{
  int row;

  if (!GTK_CLIST(prefs.directory_displayed)->selection)
    return;

  row = GPOINTER_TO_INT(GTK_CLIST(prefs.directory_displayed)->selection->data);

  gtk_clist_row_move(GTK_CLIST(prefs.directory_displayed), row, row + 1);

  on_preferences_changed();
}

void on_queue_add_clicked(GtkButton * button, gpointer data)
{
  int index, old_row, new_row;
  char *values[1];

  if (!GTK_CLIST(prefs.queue_possible)->selection)
    return;

  old_row = GPOINTER_TO_INT(GTK_CLIST(prefs.queue_possible)->selection->data);
  index =
    GPOINTER_TO_INT(gtk_clist_get_row_data
		    (GTK_CLIST(prefs.queue_possible), old_row));
  gtk_clist_get_text(GTK_CLIST(prefs.queue_possible), old_row, 0, &values[0]);

  new_row = gtk_clist_append(GTK_CLIST(prefs.queue_displayed), values);
  gtk_clist_set_row_data(GTK_CLIST(prefs.queue_displayed), new_row,
			 GINT_TO_POINTER(index));

  gtk_clist_remove(GTK_CLIST(prefs.queue_possible), old_row);

  on_preferences_changed();
}

void on_queue_remove_clicked(GtkButton * button, gpointer data)
{
  int index, old_row, new_row;
  char *values[1];

  if (!GTK_CLIST(prefs.queue_displayed)->selection)
    return;

  old_row = GPOINTER_TO_INT(GTK_CLIST(prefs.queue_displayed)->selection->data);
  index =
    GPOINTER_TO_INT(gtk_clist_get_row_data
		    (GTK_CLIST(prefs.queue_displayed), old_row));
  gtk_clist_get_text(GTK_CLIST(prefs.queue_displayed), old_row, 0, &values[0]);

  new_row = gtk_clist_insert(GTK_CLIST(prefs.queue_possible), index, values);
  gtk_clist_set_row_data(GTK_CLIST(prefs.queue_possible), new_row,
			 GINT_TO_POINTER(index));

  gtk_clist_remove(GTK_CLIST(prefs.queue_displayed), old_row);

  on_preferences_changed();
}

void on_queue_up_clicked(GtkButton * button, gpointer data)
{
  int row;

  if (!GTK_CLIST(prefs.queue_displayed)->selection)
    return;

  row = GPOINTER_TO_INT(GTK_CLIST(prefs.queue_displayed)->selection->data);
  /* Don't try to move the row if it's already at the top */
  if (row == 0)
    return;

  gtk_clist_row_move(GTK_CLIST(prefs.queue_displayed), row, row - 1);

  on_preferences_changed();
}

void on_queue_down_clicked(GtkButton * button, gpointer data)
{
  int row;

  if (!GTK_CLIST(prefs.queue_displayed)->selection)
    return;

  row = GPOINTER_TO_INT(GTK_CLIST(prefs.queue_displayed)->selection->data);

  gtk_clist_row_move(GTK_CLIST(prefs.queue_displayed), row, row + 1);

  on_preferences_changed();
}
