/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  DeadFTP
 *  Copyright (C) 2000-2001 Brandon Lees
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdarg.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pthread.h>
#include <signal.h>
#include <errno.h>
#include <ctype.h>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>
#include <glade/glade.h>
#include <gnome-xml/tree.h>
#include <gnome-xml/parser.h>

#include "globals.h"
#include "main.h"
#include "ftp.h"
#include "remote.h"
#include "socket.h"
#include "hostmanager.h"
#include "layout.h"
#include "support.h"
#include "fxp.h"

/* Callbacks in this file */
gboolean on_remote_button_press_event(GtkWidget * widget, GdkEventButton * event,
				      gpointer user_data);
void on_remote_select_row(GtkCList * clist, gint row, gint column, GdkEvent * event,
			  gpointer user_data);
void on_remote_textcombo_activate(GtkEditable * editable, gpointer user_data);
void on_remote_up_clicked(GtkButton * button, gpointer user_data);
void on_remote_unselect_row(GtkCList * clist, gint row, gint column,
			    GdkEvent * event, gpointer user_data);
void on_remote_unselect_all(GtkCList * clist, gpointer user_data);
void on_remote_drag_data_get(GtkWidget * widget, GdkDragContext * drag_context,
			     GtkSelectionData * data, guint info, guint time,
			     gpointer user_data);
gboolean on_remote_drag_drop(GtkWidget * widget, GdkDragContext * drag_context,
			     gint x, gint y, guint time);
void on_readme_ok_clicked(GtkButton * button, gpointer user_data);
void on_quick_hostname_activate(GtkEditable * editable, gpointer user_data);
void on_quick_port_activate(GtkEditable * editable, gpointer user_data);
void on_quick_username_activate(GtkEditable * editable, gpointer user_data);
void on_quick_password_activate(GtkEditable * editable, gpointer user_data);
void on_connect_clicked(GtkButton * button, gpointer user_data);
void on_disconnect_clicked(GtkButton * button, gpointer user_data);
void on_connect1_activate(GtkMenuItem * menuitem, gpointer user_data);
void on_disconnect1_activate(GtkMenuItem * menuitem, gpointer user_data);
void on_remote_click_column(GtkCList * clist, gint column);
void on_remote_arrange_name(void);
void on_remote_arrange_length(void);
void on_remote_arrange_mode(void);
void on_remote_arrange_owner(void);
void on_remote_arrange_group(void);
void on_remote_arrange_date(void);
void on_remote_arrange_ascending(void);
void on_remote_arrange_descending(void);
void on_remote_arrange_autosize(GtkMenuItem *);
void on_remote_mode(void);
void on_remote_move(void);
void on_remote_transfer(void);
void on_remote_raw_command(void);
void on_remote_edit_commands(void);
void on_remote_add_queue(void);
void on_remote_rename(void);
void on_remote_mkdir(void);
void on_remote_remove(void);
void on_remote_refresh(void);

/* Local function */
void print_trans_message(void);
gint remote_connect_time(gpointer data);
void remote_dblclick(void);
void remote_quick_connect(void);
static void remote_set_sort_info(gint, gboolean);

GList *remote_row = NULL;
GList *remote_files = NULL;
GMemChunk *afile_chunk = NULL;

void remote_init()
{
  int id;

  id = gtk_statusbar_get_context_id(GTK_STATUSBAR(deadftp.status_state), "default");
  gtk_statusbar_push(GTK_STATUSBAR(deadftp.status_state), id, _("Idle"));

  id =
    gtk_statusbar_get_context_id(GTK_STATUSBAR(deadftp.status_speed),
				 STATUS_CTX_SPEED);
  gtk_statusbar_push(GTK_STATUSBAR(deadftp.status_speed), id, "");

  id =
    gtk_statusbar_get_context_id(GTK_STATUSBAR(deadftp.status_elapsed),
				 STATUS_CTX_ELAPSED);
  gtk_statusbar_push(GTK_STATUSBAR(deadftp.status_elapsed), id, "");

  id =
    gtk_statusbar_get_context_id(GTK_STATUSBAR(deadftp.status_remaining),
				 STATUS_CTX_REMAINING);
  gtk_statusbar_push(GTK_STATUSBAR(deadftp.status_remaining), id, "");

  id =
    gtk_statusbar_get_context_id(GTK_STATUSBAR(deadftp.status_queue),
				 STATUS_CTX_QUEUE);
  gtk_statusbar_push(GTK_STATUSBAR(deadftp.status_queue), id, "");

  ftp_history_set_list(deadftp_prefs.hostname_history, deadftp.quick_hostname_combo);
  gtk_entry_set_text(GTK_ENTRY(deadftp.quick_hostname), "");

  ftp_history_set_list(deadftp_prefs.port_history, deadftp.quick_port_combo);
  gtk_entry_set_text(GTK_ENTRY(deadftp.quick_port), "21");

  ftp_history_set_list(deadftp_prefs.username_history, deadftp.quick_username_combo);
  gtk_entry_set_text(GTK_ENTRY(deadftp.quick_username), "");

  afile_chunk = g_mem_chunk_create(struct afile, 128, G_ALLOC_AND_FREE);
}

void init_remote_dir_layout()
{
  int i;
  GtkWidget *label, *box;

  gtk_clist_freeze(GTK_CLIST(deadftp.remote));

  for (i = 1; i < NUM_DIR_LAYOUTS; i++) {
    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.remote), i, FALSE);
  }

  for (i = 0; i < NUM_DIR_LAYOUTS; i++) {
    label = gtk_label_new("");
    box = gtk_hbox_new(FALSE, 0);

    gtk_widget_show(label);
    gtk_widget_show(box);

    gtk_box_pack_start_defaults(GTK_BOX(box), label);
    gtk_clist_set_column_widget(GTK_CLIST(deadftp.remote), i, box);

    deadftp.remote_boxs[i] = box;
    deadftp.remote_titles[i] = label;
  }

  if (deadftp_prefs.directory_layout[DIR_LAYOUT_FILE] != -1) {
    label = deadftp.remote_titles[deadftp_prefs.directory_layout[DIR_LAYOUT_FILE]];

    gtk_label_set(GTK_LABEL(label), _(dir_layout_titles[DIR_LAYOUT_FILE]));
    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.remote),
				    deadftp_prefs.directory_layout[DIR_LAYOUT_FILE],
				    TRUE);
  }
  if (deadftp_prefs.directory_layout[DIR_LAYOUT_LENGTH] != -1) {
    label = deadftp.remote_titles[deadftp_prefs.directory_layout[DIR_LAYOUT_LENGTH]];

    gtk_label_set(GTK_LABEL(label), _(dir_layout_titles[DIR_LAYOUT_LENGTH]));
    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.remote),
				    deadftp_prefs.
				    directory_layout[DIR_LAYOUT_LENGTH], TRUE);
  }
  if (deadftp_prefs.directory_layout[DIR_LAYOUT_MODE] != -1) {
    label = deadftp.remote_titles[deadftp_prefs.directory_layout[DIR_LAYOUT_MODE]];

    gtk_label_set(GTK_LABEL(label), _(dir_layout_titles[DIR_LAYOUT_MODE]));
    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.remote),
				    deadftp_prefs.directory_layout[DIR_LAYOUT_MODE],
				    TRUE);
  }
  if (deadftp_prefs.directory_layout[DIR_LAYOUT_OWNER] != -1) {
    label = deadftp.remote_titles[deadftp_prefs.directory_layout[DIR_LAYOUT_OWNER]];

    gtk_label_set(GTK_LABEL(label), _(dir_layout_titles[DIR_LAYOUT_OWNER]));
    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.remote),
				    deadftp_prefs.directory_layout[DIR_LAYOUT_OWNER],
				    TRUE);
  }
  if (deadftp_prefs.directory_layout[DIR_LAYOUT_GROUP] != -1) {
    label = deadftp.remote_titles[deadftp_prefs.directory_layout[DIR_LAYOUT_GROUP]];

    gtk_label_set(GTK_LABEL(label), _(dir_layout_titles[DIR_LAYOUT_GROUP]));
    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.remote),
				    deadftp_prefs.directory_layout[DIR_LAYOUT_GROUP],
				    TRUE);
  }
  if (deadftp_prefs.directory_layout[DIR_LAYOUT_DATE] != -1) {
    label = deadftp.remote_titles[deadftp_prefs.directory_layout[DIR_LAYOUT_DATE]];

    gtk_label_set(GTK_LABEL(label), _(dir_layout_titles[DIR_LAYOUT_DATE]));
    gtk_clist_set_column_visibility(GTK_CLIST(deadftp.remote),
				    deadftp_prefs.directory_layout[DIR_LAYOUT_DATE],
				    TRUE);
  }

  for (i = 0; i < NUM_DIR_LAYOUTS; i++) {
    gtk_clist_set_column_auto_resize(GTK_CLIST(deadftp.remote), i,
				     deadftp_prefs.remote_auto_size);
  }

  if (deadftp_prefs.directory_layout[deadftp_prefs.remote_sort_field] == -1) {
    for (deadftp_prefs.remote_sort_field = 0;
	 (deadftp_prefs.directory_layout[deadftp_prefs.remote_sort_field] == -1)
	 && deadftp_prefs.remote_sort_field < NUM_DIR_LAYOUTS;
	 deadftp_prefs.remote_sort_field++) {
      ;				/* do nothing */
    }
  }

  box =
    deadftp.remote_boxs[deadftp_prefs.
			directory_layout[deadftp_prefs.remote_sort_field]];
  deadftp.remote_arrow =
    load_pixmap(deadftp.window,
		(deadftp_prefs.remote_sort_order ==
		 DIR_SORT_ASC) ? "asc.xpm" : "dsc.xpm", 0);
  gtk_box_pack_end(GTK_BOX(box), deadftp.remote_arrow, FALSE, FALSE, 1);
  gtk_widget_show(deadftp.remote_arrow);

  gtk_clist_thaw(GTK_CLIST(deadftp.remote));
}

void print_trans_message()
{
  unsigned long t;
  double speed;
  char c;

  t = time(NULL);
  t -= QUEUE_CURRENT(FTP.queue)->trans_start;
  speed = (double) QUEUE_CURRENT(FTP.queue)->total_trans / (double) t;
  c = 'B';
  if (speed > ONE_KILOBYTE) {
    speed /= ONE_KILOBYTE;
    c = 'K';
  }
  if (speed > ONE_KILOBYTE) {
    speed /= ONE_KILOBYTE;
    c = 'M';
  }
  print_console(CONSOLE_INFO,
		_("Transferred file %s in %.2ld:%.2ld:%.2ld %.2f %c/sec\n"),
		QUEUE_CURRENT(FTP.queue)->local_file, t / 3600, (t % 3600) / 60,
		(t % 3600) % 60, speed, c);
}

mode_t str_to_mode(const char *mode)
{
  mode_t m = S_IRUSR | S_IWUSR;

  /* Really, the only useful information we can get from the FTP server mode is 
   * if it's an executable.
   */
  if (!mode || strlen(mode) < 9) {
    return m;
  }

  if (mode[1] == 'r') {
    m |= S_IRUSR;
  }
  if (mode[2] == 'w') {
    m |= S_IWUSR;
  }
  if (mode[3] == 'x') {
    m |= S_IXUSR;
  } else if (mode[3] == 's') {
    /* Don't set SETUID bits, just set exacutable bit */
    m |= S_IXUSR;
  }
  if (mode[4] == 'r') {
    m |= S_IRGRP;
  }
  if (mode[5] == 'w') {
    m |= S_IWGRP;
  }
  if (mode[6] == 'x') {
    m |= S_IXGRP;
  } else if (mode[6] == 's') {
    /* Don't set SETGID bits, just set exacutable bit */
    m |= S_IXGRP;
  }
  if (mode[7] == 'r') {
    m |= S_IROTH;
  }
  if (mode[8] == 'w') {
    m |= S_IWOTH;
  }
  if (mode[9] == 'x') {
    m |= S_IXOTH;
  }
  return m;
}

void remote_update_dir_list()
{
  int i, j, row, show_hidden = 0;
  char *dirent[NUM_DIR_LAYOUTS];
  struct afile *x;

  gtk_clist_freeze(GTK_CLIST(deadftp.remote));
  gtk_clist_clear(GTK_CLIST(deadftp.remote));

  if (FTP.host_entry) {
    show_hidden = hm_get_pref_int("show_hidden");
  }
  if (!FTP.host_entry || (show_hidden < 0)) {
    show_hidden = deadftp_prefs.list_method;
  }

  /* remote_files = g_list_sort(remote_files, dir_sort_funcs[deadftp_prefs.remote_sort_field].func); */
  /* The list should now be sorted as it is built */
  if (deadftp_prefs.remote_sort_order == DIR_SORT_DSC) {
    remote_files = g_list_reverse(remote_files);
  }

  for (i = 0, row = 0; i < g_list_length(remote_files); i++) {
    x = g_list_nth_data(remote_files, i);

    if (x->file && *x->file == '.' && show_hidden != LIST_SHOW_HIDDEN) {
      continue;
    }

    for (j = 0; j < NUM_DIR_LAYOUTS; j++) {
      dirent[j] = NULL;
    }

    if (deadftp_prefs.directory_layout[DIR_LAYOUT_FILE] != -1)
      dirent[deadftp_prefs.directory_layout[DIR_LAYOUT_FILE]] = x->file;
    if (deadftp_prefs.directory_layout[DIR_LAYOUT_LENGTH] != -1)
      dirent[deadftp_prefs.directory_layout[DIR_LAYOUT_LENGTH]] = x->length;
    if (deadftp_prefs.directory_layout[DIR_LAYOUT_MODE] != -1)
      dirent[deadftp_prefs.directory_layout[DIR_LAYOUT_MODE]] = x->mode;
    if (deadftp_prefs.directory_layout[DIR_LAYOUT_OWNER] != -1)
      dirent[deadftp_prefs.directory_layout[DIR_LAYOUT_OWNER]] = x->owner;
    if (deadftp_prefs.directory_layout[DIR_LAYOUT_GROUP] != -1)
      dirent[deadftp_prefs.directory_layout[DIR_LAYOUT_GROUP]] = x->group;
    if (deadftp_prefs.directory_layout[DIR_LAYOUT_DATE] != -1)
      dirent[deadftp_prefs.directory_layout[DIR_LAYOUT_DATE]] = x->date;

    gtk_clist_append(GTK_CLIST(deadftp.remote), dirent);

    gtk_clist_set_row_data(GTK_CLIST(deadftp.remote), row, x);

    set_clist_icon(deadftp.remote, row,
		   deadftp_prefs.directory_layout[DIR_LAYOUT_FILE], x->file,
		   x->is_dir);

    row++;
  }
  gtk_clist_thaw(GTK_CLIST(deadftp.remote));

}

gint remote_connect_time(gpointer data)
{
  unsigned long t;
  int len, timeout = 0, keepalive = 0;
  char buffer[SMALL_BUFFER];

  if (FTP.state == FTP_STATE_DISCONNECTED)
    return FALSE;
  t = time(NULL);
  t -= FTP.connect;
  g_snprintf(buffer, SMALL_BUFFER, _("Connected %.2ld:"), t / 3600);
  t %= 3600;
  len = strlen(buffer);
  g_snprintf(buffer + len, SMALL_BUFFER - len, "%.2ld:", t / 60);
  t %= 60;
  len = strlen(buffer);
  g_snprintf(buffer + len, SMALL_BUFFER - len, "%.2ld", t);

  /*
   * gtk_statusbar_pop(GTK_STATUSBAR(deadftp.status1), status1_id);
   * gtk_statusbar_push(GTK_STATUSBAR(deadftp.status1), status1_id, buffer);
   */

  if (FTP.host_entry) {
    timeout = hm_get_pref_int("timeout");
  }
  if (!FTP.host_entry || (timeout < 0)) {
    timeout = deadftp_prefs.connect_timeout;
  }

  t = time(NULL);
  if (((t - FTP.timeout) >= timeout)
      && (FTP.state != FTP_STATE_IDLE)) {
    print_console(CONSOLE_ERROR, _("Connection timed out after %d seconds\n"),
		  t - FTP.timeout);
    if (FTP.state == FTP_STATE_TRANS_RETR) {
      SET_FLAG(FTP_FLAG_START_TRANS);
      queue_abort();
    }
    ftp_disconnect_retry();
  }

  if (FTP.host_entry) {
    keepalive = hm_get_pref_int("keepalive");
  }
  if (!FTP.host_entry || (keepalive < 0)) {
    keepalive = deadftp_prefs.anti_idle;
  }
  if ((FTP.state == FTP_STATE_IDLE) && FTP.keepalive
      && ((t - FTP.keepalive) >= deadftp_prefs.anti_idle_time) && keepalive) {
    FTP.keepalive = time(NULL);
    ftp_send_command(FTP_STATE_IDLE, 0, "", "NOP");
  }

  return TRUE;
}

void remote_quick_connect()
{
  gchar *host, *port, *user, *pass, p[32];

  host = gtk_entry_get_text(GTK_ENTRY(deadftp.quick_hostname));
  port = gtk_entry_get_text(GTK_ENTRY(deadftp.quick_port));
  user = gtk_entry_get_text(GTK_ENTRY(deadftp.quick_username));
  pass = gtk_entry_get_text(GTK_ENTRY(deadftp.quick_password));
  strncpy(FTP.server.hostname, host, SMALL_BUFFER);
  ftp_history_add_item(&deadftp_prefs.hostname_history, g_strdup(host));
  ftp_history_set_list(deadftp_prefs.hostname_history, deadftp.quick_hostname_combo);
  gtk_entry_set_text(GTK_ENTRY(deadftp.quick_hostname), FTP.server.hostname);

  if (*user)
    strncpy(FTP.server.username, user, SMALL_BUFFER);
  else
    strncpy(FTP.server.username, "anonymous", SMALL_BUFFER);
  ftp_history_add_item(&deadftp_prefs.username_history,
		       g_strdup(FTP.server.username));
  ftp_history_set_list(deadftp_prefs.username_history, deadftp.quick_username_combo);
  gtk_entry_set_text(GTK_ENTRY(deadftp.quick_username), FTP.server.username);

  if (*pass)
    strncpy(FTP.server.password, pass, SMALL_BUFFER);
  else
    strncpy(FTP.server.password, deadftp_prefs.anonymous_password, SMALL_BUFFER);
  if (*port)
    FTP.server.port = atoi(port);
  else
    FTP.server.port = 21;
  ftp_history_add_item(&deadftp_prefs.port_history, g_strdup(port));
  ftp_history_set_list(deadftp_prefs.port_history, deadftp.quick_port_combo);
  g_snprintf(p, 32, "%d", FTP.server.port);
  gtk_entry_set_text(GTK_ENTRY(deadftp.quick_port), p);

  CLEAR_FLAG(FTP_FLAG_CHANGE_DIR);

  FTP.host_entry = NULL;

  ftp_connect();
}

void remote_dblclick()
{
  char buffer[REGULAR_BUFFER];
  struct afile *f;
  gint src;

  g_warning("remote_row->data = %d selection->data = %d\n",
	    remote_row ? GPOINTER_TO_INT(remote_row->data) : -1,
	    GTK_CLIST(deadftp.remote)->
	    selection ? GPOINTER_TO_INT(GTK_CLIST(deadftp.remote)->selection->
					data) : -1);

  if (!remote_row)
    return;

  src = GPOINTER_TO_INT(remote_row->data);
  f = (struct afile *) gtk_clist_get_row_data(GTK_CLIST(deadftp.remote), src);
  if (f->is_dir) {
    /* only add FTP.server.current_directory to the CWD command if it contains more than just / */
    ftp_send_command(FTP_STATE_CWD, FTP_CMD_FLAG_LIST, _("Chaning Directory"),
		     "CWD %s/%s", (*FTP.server.current_directory == '/'
				   && *(FTP.server.current_directory + 1) ==
				   '\0') ? "" : FTP.server.current_directory,
		     f->file);
  } else {
    getcwd(buffer, REGULAR_BUFFER);
    add_queue(&FTP.server, f, buffer, f->file, FALSE);
  }
}

void remote_change_dir(const char *dir)
{

  ftp_send_command(FTP_STATE_CWD, FTP_CMD_FLAG_LIST, _("Changing Directory"),
		   "CWD %s", dir);

  remote_change_combo(dir);
}

void remote_change_combo(const char *dir)
{
  GList *strings = NULL, *l;
  char cwd[SMALL_BUFFER], *ch;

  strncpy(cwd, dir, SMALL_BUFFER);
  ch = strrchr(cwd, '/');
  strings = g_list_append(NULL, g_strdup(cwd));
  while (ch) {
    *ch = '\0';
    strings = g_list_append(strings, g_strdup(cwd));
    ch = strrchr(cwd, '/');
  }
  gtk_combo_set_popdown_strings(GTK_COMBO(deadftp.remote_combo), strings);

  l = strings;
  while (l) {
    ch = (char *) l->data;
    g_free(ch);
    l = g_list_next(l);
  }
  g_list_free(l);
}

void on_remote_mkdir()
{
  load_newDir_dialog();

  gtk_object_set_data(GTK_OBJECT(deadftp.newDir.dialog), "isLocal",
		      GINT_TO_POINTER(0));

}

void on_remote_rename()
{
  /* TODO: have this funtion bring up dialog for each of the selected files */
  int src;
  struct afile *f;

  if (!remote_row)
    return;

  src = GPOINTER_TO_INT(remote_row->data);

  f = (struct afile *) gtk_clist_get_row_data(GTK_CLIST(deadftp.remote), src);

  load_rename_dialog();
  gtk_label_set_text(GTK_LABEL(deadftp.renameDialog.filename), f->file);
  gtk_object_set_data(GTK_OBJECT(deadftp.renameDialog.dialog), "remote_filename",
		      f->file);
}

void remote_do_rename(const char *oldfile, const char *newfile)
{
  ftp_send_command(FTP_STATE_RENAME_FROM, 0, _("Renaming File"), "RNFR %s", oldfile);
  if (FTP.rename) {
    g_free(FTP.rename);
    FTP.rename = NULL;
  }
  FTP.rename = g_strdup(newfile);
}

void on_remote_remove()
{
  gint src;
  GList *r;
  struct afile *f;

  if (!remote_row)
    return;

  for (r = remote_row; r; r = g_list_next(r)) {
    src = GPOINTER_TO_INT(r->data);

    f = (struct afile *) gtk_clist_get_row_data(GTK_CLIST(deadftp.remote), src);
    /* TODO: Need to determine if this is a directory and if so act appropriatly
     * to remove it */
    if (f->is_dir)
      ftp_send_command(FTP_STATE_DELETE, FTP_CMD_FLAG_LIST, _("Removing Directory"),
		       "RMD %s", f->file);
    else
      ftp_send_command(FTP_STATE_DELETE, FTP_CMD_FLAG_LIST, _("Removing File"),
		       "DELE %s", f->file);
  }
}
void on_remote_add_queue()
{
  char buffer[REGULAR_BUFFER];
  GList *r;
  gint src;
  struct afile *f;

  if (!remote_row)
    return;
  for (r = remote_row; r; r = g_list_next(r)) {
    src = GPOINTER_TO_INT(r->data);
    f = (struct afile *) gtk_clist_get_row_data(GTK_CLIST(deadftp.remote), src);

    getcwd(buffer, REGULAR_BUFFER);
    add_queue(&FTP.server, f, buffer, f->file, FALSE);
  }
}

gboolean on_remote_button_press_event(GtkWidget * widget, GdkEventButton * event,
				      gpointer user_data)
{
  if ((event->type == GDK_2BUTTON_PRESS) && (event->button == 1))
    remote_dblclick();

  return FALSE;
}

void on_remote_select_row(GtkCList * clist, gint row, gint column, GdkEvent * event,
			  gpointer user_data)
{
  if (!row_select_find(remote_row, row)) {
    remote_row = g_list_append(remote_row, GINT_TO_POINTER(row));
  }
}
void on_remote_textcombo_activate(GtkEditable * editable, gpointer user_data)
{
  gchar *dir = gtk_entry_get_text(GTK_ENTRY(editable));

  remote_change_dir(dir);
}

void on_remote_up_clicked(GtkButton * button, gpointer user_data)
{
  ftp_send_command(FTP_STATE_CWD, FTP_CMD_FLAG_LIST, _("Chaging Directory"), "CDUP");
}

void on_remote_unselect_row(GtkCList * clist, gint row, gint column,
			    GdkEvent * event, gpointer user_data)
{
  GList *x = row_select_find(remote_row, row);

  if (!x)
    return;
  remote_row = g_list_remove(remote_row, x->data);
}

void on_remote_unselect_all(GtkCList * clist, gpointer user_data)
{
  g_print("unselect all for remote\n");
}
static int is_special_char(char *c)
{
  /* I don't know what the real chars are, I should look it up eventually */
  char special_chars[] = "~`!@#$%^&*()-_=+'\";:/?.>,<\\|[{]}";
  int i;

  for (i = 0; i < strlen(special_chars); i++) {
    if (*c == special_chars[i]) {
      return 1;
    }
  }
  return 0;
}

void on_remote_drag_data_get(GtkWidget * widget, GdkDragContext * drag_context,
			     GtkSelectionData * data, guint info, guint time,
			     gpointer user_data)
{
  char cwd[SMALL_BUFFER], mydata[BIG_BUFFER];
  GList *r;
  gint src;
  struct afile *f;

  /* Possible FXP, check to see if it's in the list of possible targets */
  if (data->target == gdk_atom_intern("xml/fxp-info", FALSE)) {
    /* This is an FXP get an fxp_desc and send it */
    fxp_desc_t *fxp = fxp_get();
    xmlDocPtr doc = xmlNewDoc("1.0");
    xmlNodePtr node;
    xmlChar *ddoc = NULL;
    gchar intdata[SMALL_BUFFER];
    int size = 0;

    g_warning("Starting FXP setup\n");

    doc->root = xmlNewDocNode(doc, NULL, "fxpinfo", NULL);
    node = xmlNewChild(doc->root, NULL, "fxp", NULL);
    /* set a version for compatability checking */
    xmlSetProp(node, "version", "1.0");

    g_snprintf(intdata, SMALL_BUFFER, "%d", fxp->shm_id);
    xmlNewTextChild(node, NULL, "shmid", intdata);

    for (r = remote_row; r; r = g_list_next(r)) {
      gchar intdata[128];
      xmlNodePtr node = xmlNewChild(doc->root, NULL, "item", NULL);
      queueid_t qid;

      src = GPOINTER_TO_INT(r->data);
      f = (struct afile *) gtk_clist_get_row_data(GTK_CLIST(deadftp.remote), src);

      qid =
	queue_add_fxp(fxp, FTP.server.hostname, FTP.server.port, FTP.server.username,
		      FTP.server.password, FTP.server.current_directory, f->file,
		      f->length ? atoi(f->length) : 0, 0);
      xmlNewTextChild(node, NULL, "hostname", FTP.server.hostname);
      xmlNewTextChild(node, NULL, "filename", f->file);
      xmlNewTextChild(node, NULL, "length", f->length ? f->length : "0");
      g_snprintf(intdata, 128, "%ld", qid);
      xmlNewTextChild(node, NULL, "id", intdata);
    }
    xmlDocDumpMemory(doc, &ddoc, &size);
    gtk_selection_data_set(data, data->target, 8, ddoc, size + 1);

    xmlFreeDoc(doc);
  } else if (data->target == gdk_atom_intern("xml/file-info", FALSE)) {
    xmlDocPtr doc = xmlNewDoc("1.0");
    xmlChar *ddoc = NULL;
    int size = 0;
    doc->root = xmlNewDocNode(doc, NULL, "fileinfo", NULL);

    for (r = remote_row; r; r = g_list_next(r)) {
      gchar port[128];
      xmlNodePtr node = xmlNewChild(doc->root, NULL, "item", NULL);

      src = GPOINTER_TO_INT(r->data);
      f = (struct afile *) gtk_clist_get_row_data(GTK_CLIST(deadftp.remote), src);
      getcwd(cwd, SMALL_BUFFER);
      xmlNewTextChild(node, NULL, "hostname", FTP.server.hostname);

      g_snprintf(port, 128, "%d", FTP.server.port);
      xmlNewTextChild(node, NULL, "port", port);

      xmlNewTextChild(node, NULL, "username", FTP.server.username);
      xmlNewTextChild(node, NULL, "password", FTP.server.password);
      xmlNewTextChild(node, NULL, "remote_file", f->file);
      xmlNewTextChild(node, NULL, "remote_dir", FTP.server.current_directory);
      xmlNewTextChild(node, NULL, "local_file", f->file);
      xmlNewTextChild(node, NULL, "local_dir", cwd);

      xmlNewTextChild(node, NULL, "length", f->length ? f->length : "-1");
      xmlNewTextChild(node, NULL, "date", f->date);

      if (f->is_dir)
	xmlNewChild(node, NULL, "directory", NULL);
    }
    xmlDocDumpMemory(doc, &ddoc, &size);

    gtk_selection_data_set(data, data->target, 8, ddoc, size + 1);

    xmlFreeDoc(doc);
  } else {
    *mydata = '\0';
    for (r = remote_row; r; r = g_list_next(r)) {
      src = GPOINTER_TO_INT(r->data);
      f = (struct afile *) gtk_clist_get_row_data(GTK_CLIST(deadftp.remote), src);
      getcwd(cwd, SMALL_BUFFER);
      if (g_strcasecmp(FTP.server.username, "anonymous") == 0)
	g_snprintf(mydata, BIG_BUFFER, "%sftp://%s:%d%s/%s\n", mydata,
		   FTP.server.hostname, FTP.server.port,
		   FTP.server.current_directory, f->file);
      else {
	char *d, *s;
	strcat(mydata, "ftp://");
	d = mydata + strlen(mydata);
	for (s = FTP.server.username; s && *s; s++) {
	  if (is_special_char(s)) {
	    g_snprintf(mydata, BIG_BUFFER, "%s%%%X", mydata, *s & 0xff);
	    d += 3;
	  } else
	    *d++ = *s;
	}
	*d++ = ':';
	for (s = FTP.server.password; s && *s; s++) {
	  if (is_special_char(s)) {
	    g_snprintf(mydata, BIG_BUFFER, "%s%%%X", mydata, *s & 0xff);
	    d += 3;
	  } else
	    *d++ = *s;
	}
	*d++ = '@';
	g_snprintf(d, BIG_BUFFER - (d - mydata), "%s:%d%s/%s\n", FTP.server.hostname,
		   FTP.server.port, FTP.server.current_directory, f->file);
      }
    }
    gtk_selection_data_set(data, data->target, 8, mydata, strlen(mydata) + 1);
  }
}

void on_readme_ok_clicked(GtkButton * button, gpointer user_data)
{
  gtk_widget_destroy(deadftp.showReadme);
  deadftp.showReadme = NULL;
}

void on_quick_hostname_activate(GtkEditable * editable, gpointer user_data)
{
  remote_quick_connect();
}

void on_quick_port_activate(GtkEditable * editable, gpointer user_data)
{
  remote_quick_connect();
}

void on_quick_username_activate(GtkEditable * editable, gpointer user_data)
{
  remote_quick_connect();
}

void on_quick_password_activate(GtkEditable * editable, gpointer user_data)
{
  remote_quick_connect();
}

void on_connect_clicked(GtkButton * button, gpointer user_data)
{
  remote_quick_connect();
}

void on_disconnect_clicked(GtkButton * button, gpointer user_data)
{
  ftp_disconnect();

  if (FTP.retry_tag != -1) {
    gtk_timeout_remove(FTP.retry_tag);
    print_console(CONSOLE_INFO, _("Retry Cancelled\n"));
    FTP.retry_tag = -1;
  }
}

void on_connect1_activate(GtkMenuItem * menuitem, gpointer user_data)
{
  remote_quick_connect();
}

void on_disconnect1_activate(GtkMenuItem * menuitem, gpointer user_data)
{
  ftp_disconnect();
}

void on_remote_click_column(GtkCList * clist, gint column)
{
  int field = 0;

  while ((field < NUM_DIR_LAYOUTS)
	 && (deadftp_prefs.directory_layout[field] != column))
    field++;

  g_return_if_fail((field >= 0) && (field < NUM_DIR_LAYOUTS));

  if (field == deadftp_prefs.remote_sort_field) {
    remote_set_sort_info(deadftp_prefs.remote_sort_field,
			 !deadftp_prefs.remote_sort_order);
  } else {
    remote_set_sort_info(field, DIR_SORT_ASC);
  }
}

void ftp_history_add_item(GList ** history, char *value)
{
  GList *l;

  for (l = *history; l; l = l->next) {
    char *val2;

    val2 = (char *) l->data;
    if (g_strcasecmp(value, val2) == 0) {
      g_free(value);
      return;
    }
  }
  if (g_list_length(*history) >= FTP_HISTORY_SAVE) {
    gpointer data;

    data = g_list_nth_data(*history, FTP_HISTORY_SAVE - 1);
    *history = g_list_remove(*history, data);
    g_free(data);
  }
  *history = g_list_prepend(*history, value);
}

void ftp_history_set_list(GList * history, GtkWidget * combo)
{
  GList *copy = g_list_copy(history);

  if (!copy) {
    return;
  }
  gtk_combo_set_popdown_strings(GTK_COMBO(combo), copy);
}

gboolean on_remote_drag_drop(GtkWidget * widget, GdkDragContext * drag_context,
			     gint x, gint y, guint time)
{
  /* Check for an FXP */
  if (drag_context->source_window != drag_context->dest_window) {
    GList *target = drag_context->targets;
    GdkAtom atom = gdk_atom_intern("xml/fxp-info", FALSE);

    while (target && (atom != GPOINTER_TO_INT(target->data)))
      target = g_list_next(target);

    /* It's an FXP so get eh right target data and return */
    if (target) {
      gtk_drag_get_data(widget, drag_context, atom, time);
      return TRUE;
    }
  }
  /* not handled, return FALSE so the default action is triggered */
  return FALSE;
}

void on_remote_refresh()
{
  if (FTP.state != FTP_STATE_DISCONNECTED) {
    remote_change_dir(FTP.server.current_directory);
  }
}

static void remote_set_sort_info(gint field, gboolean order)
{
  GtkWidget *box;

  /* If there is no change, don't do anything */
  if ((deadftp_prefs.remote_sort_field == field)
      && (deadftp_prefs.remote_sort_order == order))
    return;

  box =
    deadftp.remote_boxs[deadftp_prefs.
			directory_layout[deadftp_prefs.remote_sort_field]];
  gtk_container_remove(GTK_CONTAINER(box), deadftp.remote_arrow);

  deadftp_prefs.remote_sort_field = field;
  deadftp_prefs.remote_sort_order = order;

  remote_files =
    g_list_sort(remote_files, dir_sort_funcs[deadftp_prefs.remote_sort_field].func);
  /* does the old widget need to be detroyed or is that handled by the refcount?? */
  deadftp.remote_arrow =
    load_pixmap(deadftp.window, (order == DIR_SORT_ASC) ? "asc.xpm" : "dsc.xpm", 0);
  gtk_widget_show(deadftp.remote_arrow);

  box =
    deadftp.remote_boxs[deadftp_prefs.
			directory_layout[deadftp_prefs.remote_sort_field]];
  gtk_box_pack_end(GTK_BOX(box), deadftp.remote_arrow, FALSE, FALSE, 1);

  remote_update_dir_list();
}

void on_remote_arrange_name()
{
  if (deadftp_prefs.directory_layout[DIR_LAYOUT_FILE] != -1)
    remote_set_sort_info(DIR_LAYOUT_FILE, deadftp_prefs.remote_sort_order);
}

void on_remote_arrange_owner()
{
  if (deadftp_prefs.directory_layout[DIR_LAYOUT_OWNER] != -1)
    remote_set_sort_info(DIR_LAYOUT_OWNER, deadftp_prefs.remote_sort_order);
}

void on_remote_arrange_group()
{
  if (deadftp_prefs.directory_layout[DIR_LAYOUT_GROUP] != -1)
    remote_set_sort_info(DIR_LAYOUT_GROUP, deadftp_prefs.remote_sort_order);
}

void on_remote_arrange_mode()
{
  if (deadftp_prefs.directory_layout[DIR_LAYOUT_MODE] != -1)
    remote_set_sort_info(DIR_LAYOUT_MODE, deadftp_prefs.remote_sort_order);
}

void on_remote_arrange_date()
{
  if (deadftp_prefs.directory_layout[DIR_LAYOUT_DATE] != -1)
    remote_set_sort_info(DIR_LAYOUT_DATE, deadftp_prefs.remote_sort_order);
}

void on_remote_arrange_length()
{
  if (deadftp_prefs.directory_layout[DIR_LAYOUT_LENGTH] != -1)
    remote_set_sort_info(DIR_LAYOUT_LENGTH, deadftp_prefs.remote_sort_order);
}

void on_remote_arrange_ascending()
{
  remote_set_sort_info(deadftp_prefs.remote_sort_field, DIR_SORT_ASC);
}

void on_remote_arrange_descending()
{
  remote_set_sort_info(deadftp_prefs.remote_sort_field, DIR_SORT_DSC);
}

void on_remote_arrange_autosize(GtkMenuItem * item)
{
  int i;
  gboolean active = GTK_CHECK_MENU_ITEM(item)->active;

  /* don't do anything if there's no change */
  if (active == deadftp_prefs.remote_auto_size)
    return;

  deadftp_prefs.remote_auto_size = active;
  for (i = 0; i < NUM_DIR_LAYOUTS; i++)
    gtk_clist_set_column_auto_resize(GTK_CLIST(deadftp.remote), i,
				     deadftp_prefs.remote_auto_size);
  remote_update_dir_list();
}

void on_remote_mode()
{
  load_mode_dialog(FALSE);
}

void on_remote_move()
{
}

void on_remote_transfer()
{
}

void on_remote_raw_command()
{
}

void on_remote_edit_commands()
{
}

void remote_update_quick_connect()
{
  char buffer[16];
  gtk_entry_set_text(GTK_ENTRY(deadftp.quick_hostname), FTP.server.hostname);
  g_snprintf(buffer, 16, "%d", FTP.server.port);
  gtk_entry_set_text(GTK_ENTRY(deadftp.quick_port), buffer);
  gtk_entry_set_text(GTK_ENTRY(deadftp.quick_username), FTP.server.username);
  gtk_entry_set_text(GTK_ENTRY(deadftp.quick_password), FTP.server.password);
}
